/*
* SACD Decoder plugin
* Copyright (c) 2011-2016 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "../ATLHelpers/ATLHelpers.h"
#include "sacd_setup.h"
#include "proxy_output.h"

static const GUID  proxy_output_guid = { 0xf4d74794, 0x2936, 0x4595, { 0x8d, 0x6a, 0xc0, 0xce, 0x7c, 0x5b, 0x35, 0x25 } };
static const char* proxy_output_name = "DSD";

static const GUID  NULL_output_guid = { 0xEEEB07DE, 0xC2C8, 0x44C2, { 0x98, 0x5C, 0xC8, 0x58, 0x56, 0xD9, 0x6D, 0xA1 } };
static const GUID  DS_output_guid   = { 0xD41D2423, 0xFBB0, 0x4635, { 0xB2, 0x33, 0x70, 0x54, 0xF7, 0x98, 0x14, 0xAB } };

static const GUID no_proxy_guids[] = { NULL_output_guid, DS_output_guid };

const uint8_t proxy_output_t::DoP_MARKER[2] = { 0x05, 0xFA };

static GUID xor_guids(const GUID& p_guid1, const GUID& p_guid2) {
	GUID guid_xor;
	uint8_t* p_1 = (uint8_t*)&p_guid1;
	uint8_t* p_2 = (uint8_t*)&p_guid2;
	uint8_t* p_x = (uint8_t*)&guid_xor;
	for (int i = 0; i < sizeof(GUID); i++) {
		p_x[i] = p_1[i] ^ p_2[i];
	}
	return guid_xor;
}

bool proxy_output_t::g_advanced_settings_query() {
	return false;
}

bool proxy_output_t::g_needs_bitdepth_config() {
	return true;
}

bool proxy_output_t::g_needs_dither_config() {
	return true;
}

bool proxy_output_t::g_needs_device_list_prefixes() {
	return false;
}

void proxy_output_t::g_enum_devices(output_device_enum_callback& p_callback) {
	class dsd_output_device_enum_callback : public output_device_enum_callback {
		output_device_enum_callback* m_callback;
		GUID m_guid;
		const char* m_name;
	public:
		dsd_output_device_enum_callback(output_device_enum_callback* p_callback, const GUID& p_guid, const char* p_name) {
			m_callback = p_callback;
			m_guid = p_guid;
			m_name = p_name;
		}
		void on_device(const GUID& p_guid, const char* p_name, unsigned p_name_length) {
			pfc::string8 name = m_name;
			name.get_length() > 0 ? name << " : " << p_name : name << p_name;
			GUID guid_dev = xor_guids(m_guid, p_guid);
			m_callback->on_device(guid_dev, name, name.get_length());
		}
	};
	service_enum_t<output_entry> output_enum;
	service_ptr_t<output_entry> output_ptr;
	while (output_enum.next(output_ptr)) {
		GUID guid = output_ptr->get_guid();
		const char* name = output_ptr->get_name();
		if (guid != g_get_guid()) {
			bool skip_guid = false;
			for (GUID no_proxy_guid : no_proxy_guids) {
				if (no_proxy_guid == guid) {
					skip_guid = true;
				}
			}
			if (!skip_guid) {
				dsd_output_device_enum_callback dsd_callback(&p_callback, guid, name);
				output_ptr->enum_devices(dsd_callback);
			}
		}
	}
}

GUID proxy_output_t::g_get_guid() {
	return proxy_output_guid;
}

const char* proxy_output_t::g_get_name() {
	return proxy_output_name;
}

void proxy_output_t::g_advanced_settings_popup(HWND p_parent, POINT p_menupoint) {
}

bool proxy_output_t::g_get_output(service_ptr_t<output_entry>& p_output, GUID& p_guid, pfc::string8& p_name) {
	bool ok = false;
	outputCoreConfig_t output_config;
	static_api_ptr_t<output_manager> output_manager_ptr;
	output_manager_ptr->getCoreConfig(output_config);
	service_enum_t<output_entry> output_enum;
	service_ptr_t<output_entry> output_ptr;
	while (output_enum.next(output_ptr)) {
		pfc::string8 output_name;
		if (output_ptr->get_guid() != g_get_guid()) {
			GUID output_guid = xor_guids(output_ptr->get_guid(), output_config.m_device);
			if (output_ptr->get_device_name(output_guid, output_name)) {
				p_output = output_ptr;
				p_guid = output_guid;
				p_name = output_name;
				ok = true;
				break;
			}
		}
	}
	return ok;
}

proxy_output_t::proxy_output_t(const GUID& p_device, double p_buffer_length, bool p_dither, t_uint32 p_bitdepth) {
	m_volume_dB = 0;
	m_trace = CSACDPreferences::g_get_trace();
	m_paused = false;
	m_is_output_v2 = false;
	m_track_marks = false;
	m_duration_difference = 0;
	service_ptr_t<output_entry> output_ptr;
	GUID output_guid;
	pfc::string8 output_name;
	if (g_get_output(output_ptr, output_guid, output_name)) {
		output_ptr->instantiate(m_output, output_guid, p_buffer_length, p_dither, p_bitdepth);
	}
	if (m_output.is_empty()) {
		pfc::throw_exception_with_message<exception_io>("proxy_output::proxy_output() => DSD output device is not defined");
	}
	m_is_output_v2 = m_output->service_query_t(m_output_v2);
	if (m_trace) {
		console::printf(
			"proxy_output::proxy_output() => output%s [device = \"%s : %s\", buffer_length = %s, dither = %s, bitdepth = %d]",
			m_is_output_v2 ? "_v2" : "",
			output_ptr->get_name(),
			output_name.c_str(),
			pfc::format_float(p_buffer_length, 0, 5).toString(),
			p_dither ? "true" : "false",
			p_bitdepth
		);
	}
	service_enum_t<dsd_processor_service> dsddsp_enum;
	service_ptr_t<dsd_processor_service> dsddsp_temp;
	while (dsddsp_enum.next(dsddsp_temp)) {
		if (dsddsp_temp->get_guid() == CSACDPreferences::get_dsd_processor()) {
			m_dsddsp = dsddsp_temp;
			if (m_trace) {
				console::printf("proxy_output::proxy_output() => Use DSD Processor [name = \"%s\"]", m_dsddsp->get_name());
			}
			break;
		}
	}
	m_dsd_stream->set_accept_data(true);
}

proxy_output_t::~proxy_output_t() {
	m_dsd_stream->set_accept_data(false);
	m_dsd_stream->flush();
	if (m_dsddsp.is_valid()) {
		m_dsddsp->stop();
		if (m_trace) {
			console::printf("proxy_output::~proxy_output() => Stop DSD Processor");
		}
	}
	if (m_trace) {
		console::printf("proxy_output::~proxy_output()");
	}
}

double proxy_output_t::get_latency() {
	double latency = m_output->get_latency();
	/*
	if (m_trace) {
		console::printf("proxy_output::get_latency() => %s", format_float(latency, 0, 5).toString());
	}
	*/
	return latency;
}

void proxy_output_t::process_samples(const audio_chunk& p_chunk) {
	t_samplespec pcm_spec(p_chunk);
	if (!pcm_spec.is_valid()) {
		pfc::throw_exception_with_message<exception_io_data>("proxy_output::process_samples() => Invalid audio stream specifications");
	}
	dsd_chunk_t* dsd_chunk = m_dsd_stream->get_first_chunk_ptr();
	if (dsd_chunk) {
		m_duration_difference += p_chunk.get_duration();
		while (dsd_chunk && m_duration_difference > 0) {
			if (pick_dsd_processor(dsd_chunk, pcm_spec)) {
				m_dsddsp->run(dsd_chunk->get_data(), dsd_chunk->get_sample_count(), m_dsddsp_out_data);
				pack_dsd_to_dop(m_dsddsp_out_data, m_dsddsp_out_spec);
			}
			else {
				pack_dsd_to_dop(dsd_chunk->get_buffer(), dsd_chunk->get_spec());
			}
			m_output->process_samples(m_dop_chunk);
			m_duration_difference -= dsd_chunk->get_duration();
			bool output_ready = false;
			int waits = 0;
			while (!m_paused && !output_ready && m_duration_difference > 0) {
				m_output->update(output_ready);
				/*
				if (m_trace) {
					console::printf("proxy_output::process_samples() => Waiting for driver [waits = %d, output_ready = %s]", waits, output_ready ? "true" : "false");
				}
				*/
				if (!output_ready) {
					Sleep(static_cast<DWORD>(500 * dsd_chunk->get_duration()));
				}
				waits++;
				if (waits > 150) {
					pfc::throw_exception_with_message<exception_io_data>("proxy_output::process_samples() => Output ready wait time exceeded");
				}
			}
			m_dsd_stream->remove_first_chunk();
			dsd_chunk = m_dsd_stream->get_first_chunk_ptr();
		}
	}
	else {
		if (pick_dsd_processor(nullptr, pcm_spec)) {
			m_dsddsp->run(p_chunk.get_data(), p_chunk.get_sample_count(), m_dsddsp_out_data);
			pack_dsd_to_dop(m_dsddsp_out_data, m_dsddsp_out_spec);
			m_output->process_samples(m_dop_chunk);
		}
		else {
			m_output->process_samples(p_chunk);
		}
	}
	/*
	if (m_trace) {
		console::printf("proxy_output::process_samples(channels = %d, sample_rate = %d, channel_config = 0x%08x, chunk_duration = %s)",
			p_chunk.get_channels(), p_chunk.get_sample_rate(), m_spec.m_channel_config, pfc::format_float(p_chunk.get_duration(), 0, 5).toString()
		);
	}
	*/
}

void proxy_output_t::update(bool& p_ready) {
	m_output->update(p_ready);
	/*
	if (m_trace) {
		console::printf("proxy_output::update(%s)", p_ready ? "true" : "false");
	}
	*/
}

void proxy_output_t::pause(bool p_state) {
	m_output->pause(p_state);
	m_paused = p_state;
	if (m_trace) {
		console::printf("proxy_output::pause(%s)", p_state ? "true" : "false");
	}
}

void proxy_output_t::flush() {
	m_dsd_stream->flush();
	m_output->flush();
	if (m_trace) {
		console::printf("proxy_output::flush()");
	}
}

void proxy_output_t::flush_changing_track() {
	m_dsd_stream->flush();
	if (m_is_output_v2) {
		m_output_v2->flush_changing_track();
	}
	else {
		m_output->flush();
	}
	if (m_trace) {
		console::printf("proxy_output::flush_changing_track()");
	}
}

void proxy_output_t::force_play() {
	m_output->force_play();
	if (m_trace) {
		console::printf("proxy_output::force_play()");
	}
}

void proxy_output_t::volume_set(double p_val_dB) {
	m_volume_dB = p_val_dB;
	volume_adjust();
	if (m_trace) {
		console::printf("proxy_output::volume_set(%s)", pfc::format_float(p_val_dB, 0, 5).toString());
	}
}

bool proxy_output_t::want_track_marks() {
	m_track_marks = m_is_output_v2 ? m_output_v2->want_track_marks() : false;
	if (m_trace) {
		console::printf("proxy_output::want_track_marks() => %s", m_track_marks ? "true" : "false");
	}
	return true;
}

void proxy_output_t::on_track_mark() {
	if (m_track_marks) {
		m_output_v2->on_track_mark();
	}
	if (m_trace) {
		console::printf("proxy_output::on_track_mark()");
	}
}

void proxy_output_t::enable_fading(bool p_state) {
	if (m_is_output_v2) {
		m_output_v2->enable_fading(p_state);
	}
	if (m_trace) {
		console::printf("proxy_output::enable_fading(%s)", p_state ? "true" : "false");
	}
}

bool proxy_output_t::pick_dsd_processor(dsd_chunk_t* p_dsd_chunk, t_samplespec& p_pcm_spec) {
	t_samplespec inp_spec;
	t_samplespec out_spec;
	if (p_dsd_chunk) {
		inp_spec = p_dsd_chunk->get_spec();
		out_spec = p_pcm_spec;
		out_spec.m_sample_rate = p_dsd_chunk->get_spec().m_sample_rate;
	}
	else {
		inp_spec = p_pcm_spec;
		out_spec = p_pcm_spec;
	}
	if (m_dsddsp.is_valid() && m_dsddsp->is_active()) {
		out_spec.m_sample_rate = m_dsddsp_out_spec.m_sample_rate;
	}
	if (m_dsddsp.is_valid()) {
		if (m_dsddsp->is_changed() || inp_spec != m_inp_spec || out_spec != m_out_spec) {
			m_dsddsp_out_spec = out_spec;
			bool ok = m_dsddsp->start(inp_spec.m_channels, inp_spec.m_sample_rate, inp_spec.m_channel_config, m_dsddsp_out_spec.m_channels, m_dsddsp_out_spec.m_sample_rate, m_dsddsp_out_spec.m_channel_config);
			if (m_trace) {
				console::printf("proxy_output::process_samples() => Start DSD Processor [channels: %d -> %d, samplerate: %d -> %d, channel_config: 0x%08u -> 0x%08u] %s",
					inp_spec.m_channels, m_dsddsp_out_spec.m_channels,
					inp_spec.m_sample_rate, m_dsddsp_out_spec.m_sample_rate,
					inp_spec.m_channel_config, m_dsddsp_out_spec.m_channel_config,
					ok ? "running" : "not running"
				);
				m_dsddsp->set_volume(m_volume_dB);
				if (m_dsddsp->is_active()) {
					out_spec.m_sample_rate = m_dsddsp_out_spec.m_sample_rate;
				}
			}
		}
	}
	if (inp_spec != m_inp_spec) {
		m_inp_spec = inp_spec;
		if (m_trace) {
			console::printf("proxy_output::process_samples() => Input stream [channels: %d, samplerate: %d, channel_config: 0x%08u]",
				inp_spec.m_channels, inp_spec.m_sample_rate, inp_spec.m_channel_config
			);
		}
	}
	if (out_spec != m_out_spec) {
		m_out_spec = out_spec;
		volume_adjust();
		if (m_trace) {
			console::printf("proxy_output::process_samples() => Output stream [channels: %d, samplerate: %d, channel_config: 0x%08u]",
				out_spec.m_channels, out_spec.m_sample_rate, out_spec.m_channel_config
			);
		}
	}
	return m_dsddsp.is_valid() && m_dsddsp->is_active();
}

void proxy_output_t::pack_dsd_to_dop(const pfc::array_t<t_uint8>& p_inp_data, const t_samplespec& p_inp_spec) {
	unsigned out_samplerate = p_inp_spec.m_sample_rate / 16;
	t_size inp_channels = p_inp_spec.m_channels;
	t_size inp_samples = p_inp_data.get_count() / inp_channels;
	t_size out_channels = m_out_spec.m_channels;
	t_size out_samples = inp_samples / 2;
	t_size out_bytes = 3 * out_samples * out_channels;
	m_dop_data.set_size(out_bytes);
	uint8_t* dop_data = m_dop_data.get_ptr();
	t_size o = 0;
	if (m_dop_marker_n.get_count() != out_channels) {
		m_dop_marker_n.set_size(out_channels);
		for (t_size ch = 0; ch < m_dop_marker_n.get_size(); ch++) {
			m_dop_marker_n[ch] = 0;
		}
	}
	for (t_size sample = 0; sample < out_samples; sample++) {
		for (t_size ch = 0; ch < out_channels; ch++) {
			t_uint8 b_msb, b_lsb;
			b_msb = p_inp_data[(2 * sample + 0) * inp_channels + ch % inp_channels];
			b_lsb = p_inp_data[(2 * sample + 1) * inp_channels + ch % inp_channels];
			dop_data[o + 0] = b_lsb;
			dop_data[o + 1] = b_msb;
			dop_data[o + 2] = DoP_MARKER[m_dop_marker_n[ch]];
			o += 3;
			m_dop_marker_n[ch] = ++m_dop_marker_n[ch] & 1;
		}
	}
	m_dop_chunk.set_data_fixedpoint(dop_data, out_bytes, out_samplerate, out_channels, 24, m_out_spec.m_channel_config);
}

void proxy_output_t::volume_adjust() {
	m_output->volume_set((m_out_spec.m_sample_rate < 2822400) ? m_volume_dB : 0);
	if (m_dsddsp.is_valid()) {
		m_dsddsp->set_volume(m_volume_dB);
	}
}

static output_factory_t<proxy_output_t> g_proxy_output_factory;
